from typing import Dict, List, Optional, Union

from django import template
from django.core.handlers.wsgi import WSGIRequest
from django.http import Http404
from django.template.base import Node
from django.template.exceptions import TemplateSyntaxError
from django.template.loader_tags import construct_relative_path
from django.utils.safestring import SafeString, mark_safe

from coltrane.renderer import MarkdownRenderer
from coltrane.retriever import get_content_directory, get_content_paths

register = template.Library()


class NoParentError(Exception):
    pass


def _is_content_slug_in_string(content_slug: str, slugs: Optional[str]) -> bool:
    """
    Whether a content slug is included in a string. Handles if `string` is
    comma-delimited list of slugs. Also handles any individual slug
    to check having a forward-slash prefix.
    """

    if not slugs:
        return False

    if not isinstance(slugs, str):
        raise TypeError("Slugs must be a string")

    split_slugs = slugs.split(",")

    for slug in split_slugs:
        slug_to_check = slug.strip()

        if slug_to_check.startswith("/"):
            slug_to_check = slug_to_check[1:]

        if slug_to_check == content_slug:
            return True

    return False


@register.simple_tag(takes_context=True)
def directory_contents(
    context, directory: Optional[str] = None, exclude: Optional[str] = None, order_by=None
) -> List[Dict[str, str]]:
    """
    Returns a list of content metadata for a particular directory. Useful for
    listing links to content.
    """

    if not directory:
        request = context["request"]
        directory = request.path
    elif isinstance(directory, SafeString):
        # Force SafeString to be a normal string so it can be used with `Path` later
        directory = directory + ""

    if directory and directory.startswith("/"):
        directory = directory[1:]

    content_paths = get_content_paths(str(directory))
    contents = []

    for path in content_paths:
        if path.name != "index.md":
            path_slug = path.name[:-3]
            content_slug = path_slug
            content_directory = get_content_directory()

            if directory:
                content_slug = f"{directory}/{path_slug}"
                content_directory = content_directory / directory

            if _is_content_slug_in_string(content_slug=content_slug, slugs=exclude):
                continue

            (_, metadata) = MarkdownRenderer.instance().get_html_and_markdown(content_slug)

            contents.append(metadata)

    if order_by and contents:
        is_reverse = False

        if order_by[0] == "-":
            is_reverse = order_by[0] == "-"
            order_by = order_by[1:]

        def _directory_content_sorter(_metadata: Dict) -> str:
            value = _metadata.get(order_by, "") or ""

            return str(value)

        contents.sort(key=_directory_content_sorter, reverse=is_reverse)

    return contents


@register.filter()
def parent(path: Union[str, WSGIRequest] = "") -> str:
    """
    Gets the the directory above `path`.
    """

    if isinstance(path, WSGIRequest) or hasattr(path, "path"):
        # Handle if a `request` is passed in
        path = path.path

    path = path.strip()

    if path.endswith("/"):
        path = path[:-1]

    if path == "":
        raise NoParentError()

    last_slash_index = path.rindex("/")
    path = path[:last_slash_index]

    return path


class IncludeMarkdownNode(Node):
    """
    Based on: `django.template.loader_tags.IncludeNode`.
    """

    context_key = "__include_markdown_context"

    def __init__(self, template, *args, **kwargs):
        self.template = template
        super().__init__(*args, **kwargs)

    def render(self, context):
        """
        Render the specified template and context.
        """

        template_name = self.template.resolve(context)

        cache = context.render_context.dicts[0].setdefault(self, {})
        template = cache.get(template_name)

        if template is None:
            template = context.template.engine.select_template((template_name,))
            cache[template_name] = template

        (html, metadata) = MarkdownRenderer.instance().render_markdown_path(template.origin.name)

        for c in context:
            for key, value in c.items():
                if key not in metadata:
                    metadata[key] = value

        return MarkdownRenderer.instance().render_html_with_django(html, metadata)


@register.tag("include_md")
def do_include_md(parser, token):
    """
    Load a markdown template and render it with the current context.

    Based on: `django.template.loader_tags.do_include`.

    Example:
        {% include "foo/some_include" %}
    """
    bits = token.split_contents()

    if len(bits) < 2:  # noqa: PLR2004
        raise TemplateSyntaxError(
            "%r tag takes at least one argument: the name of the template to be included." % bits[0]
        )

    bits[1] = construct_relative_path(parser.origin.template_name, bits[1])

    return IncludeMarkdownNode(
        parser.compile_filter(bits[1]),
    )


@register.filter(takes_context=True)
def to_html(context: dict, text: str) -> str:
    """
    Converts markdown to HTML.
    """

    (html, metadata) = MarkdownRenderer.instance().render_markdown_text(text)
    rendered_html = MarkdownRenderer.instance().render_html_with_django(html, metadata, request=context["request"])

    return mark_safe(rendered_html)  # noqa: S308


@register.simple_tag
def raise_404(message: Optional[str] = None):
    """Raise a 404 with an optional message."""

    if message:
        raise Http404(message)

    raise Http404()


@register.simple_tag(takes_context=True)
def last_path(context: dict) -> str:
    """Return the last part of the `HTTPRequest` path.

    For example, if `request.path` is "/something/cool", "cool" would be returned.
    """

    request = context["request"]
    path = request.path_info.strip()

    if path.endswith("/"):
        path = path[:-1]

    return path.split("/")[-1:][0]


@register.simple_tag(takes_context=True)
def paths(context: dict) -> List[str]:
    """Return all parts of the `HTTPRequest` path.

    For example, if `request.path` is "/something/cool", ["something", "cool"] would be returned.
    """

    request = context["request"]
    path = request.path_info.strip()

    if path.startswith("/"):
        path = path[1:]

    if path.endswith("/"):
        path = path[:-1]

    _paths = path.split("/")

    if _paths == [""]:
        _paths = []

    return _paths



import logging
from dataclasses import dataclass
from typing import Dict, List, Tuple

from django.http import FileResponse, Http404, HttpRequest, HttpResponse
from django.shortcuts import render
from django.template import TemplateDoesNotExist
from django.template.loader import select_template
from django.utils.cache import patch_response_headers
from django.utils.timezone import now

from coltrane.config.cache import ViewCache
from coltrane.config.paths import get_file_path
from coltrane.config.settings import get_disable_wildcard_templates
from coltrane.renderer import MarkdownRenderer
from coltrane.retriever import get_data

logger = logging.getLogger(__name__)


@dataclass
class PathRanking:
    """Store a path with its score based on where the wildcard is in the path."""

    path: str

    def __init__(self, path: str):
        self.path = path
        self.score = self.score_path()

    def score_path(self):
        total_score = 0
        path = self.path.replace(".html", "")

        for idx, path_piece in enumerate(path.split("/")):
            path_idx = len(path.split("/")) - idx

            if path_piece == "*":
                # Inflate wildcards to ensure they are sorted as expected
                total_score += path_idx * 100
            else:
                total_score += path_idx

        return total_score

    def __str__(self):
        return f"{self.path} ({self.score})"


def _normalize_slug(slug: str) -> str:
    if slug is None:
        slug = ""

    if slug.endswith("/"):
        slug = slug[:-1]

    if slug.startswith("/"):
        slug = slug[1:]

    if slug == "":
        slug = "index"

    return slug


def _get_from_cache_if_enabled(slug: str) -> Tuple[str, Dict]:
    """
    Gets the slug from the cache if it's enabled.
    """

    template: str = ""
    context: Dict = {}
    view_cache = ViewCache()

    if view_cache.is_enabled:
        cache_key = f"{view_cache.cache_key_namespace}{slug}"
        cached_value = view_cache.cache.get(cache_key)

        if cached_value:
            (template, context) = cached_value

    return (template, context)


def _set_in_cache_if_enabled(slug: str, template: str, context: Dict) -> None:
    """
    Sets everything in the cache if it's enabled.
    """

    view_cache = ViewCache()

    if view_cache.is_enabled:
        cache_key = f"{view_cache.cache_key_namespace}{slug}"
        view_cache.cache.set(
            cache_key,
            (template, context),
            view_cache.seconds,
        )


def _sort_potential_templates(template_paths):
    """Sort template paths based on where the wildcard is in the directory."""

    rankings = []

    for path in template_paths:
        rankings.append(PathRanking(path))

    rankings = sorted(rankings, key=lambda r: r.score, reverse=False)

    return [r.path for r in rankings]


def _get_potential_wildcard_templates(slug: str) -> List[str]:
    """Get a list of potential wildcard HTML templates based on the slug."""

    wildcard_paths = []

    slug_pieces = slug.split("/")
    slug_pieces_count = len(slug_pieces)

    for outer_idx, _ in enumerate(slug_pieces):
        for inner_idx, _ in enumerate(slug_pieces):
            new_slug_pieces = []

            if outer_idx == inner_idx:
                new_slug_pieces.extend(slug_pieces[:outer_idx])
                new_slug_pieces.extend("*")
                new_slug_pieces.extend(slug_pieces[outer_idx + 1 :])

                wildcard_paths.append(new_slug_pieces)
            elif outer_idx < inner_idx and outer_idx > 0:
                new_slug_pieces.extend(slug_pieces[:outer_idx])
                new_slug_pieces.extend("*" * (slug_pieces_count - outer_idx))

                wildcard_paths.append(new_slug_pieces)
            elif outer_idx > inner_idx and (slug_pieces_count - outer_idx - 1) > 0:
                new_slug_pieces.extend("*" * (slug_pieces_count - outer_idx - 1))
                new_slug_pieces.extend(slug_pieces[outer_idx : outer_idx + 1])
                new_slug_pieces.extend("*" * (slug_pieces_count - outer_idx - 1))

                wildcard_paths.append(new_slug_pieces)

    # Add a catch-all for everything
    # Not needed if there are no potential sub-directories
    if slug_pieces_count > 1:
        new_slug_pieces = []

        for _ in range(slug_pieces_count):
            new_slug_pieces.append("*")

        wildcard_paths.insert(0, new_slug_pieces)

    potential_templates = []

    #  Convert the arrays of paths to paths and add to the list of potential templates
    for wildcard_option in wildcard_paths:
        wildcard_option_path = "/".join(wildcard_option)
        wildcard_option_path = f"{wildcard_option_path}.html"

        potential_templates.append(wildcard_option_path)

    potential_templates = _sort_potential_templates(potential_templates)

    return potential_templates


def content(request: HttpRequest, slug: str = "index") -> HttpResponse:
    """
    Renders the markdown file stored in `content` or HTML template based on the slug from the URL.
    Adds data into the context from `data.json` and JSON files in the `data` directory.

    Will cache the rendered content if enabled.
    """

    template: str = ""
    context: Dict = {}
    slug = _normalize_slug(slug)
    slug_with_index = f"{slug}/index"

    (template, context) = _get_from_cache_if_enabled(slug)
    set_in_cache = False

    try:
        if not template or not context:
            set_in_cache = True

            (template, context) = MarkdownRenderer.instance().render_markdown(slug, request=request)
    except FileNotFoundError:
        try:
            (template, context) = MarkdownRenderer.instance().render_markdown(slug_with_index, request=request)
        except FileNotFoundError:
            # Typical templates based on the slug
            potential_templates = [
                f"{slug}.html",
                f"{slug_with_index}.html",
            ]

            if get_disable_wildcard_templates() is False:
                potential_templates.extend(_get_potential_wildcard_templates(slug))

            try:
                found_template = select_template(potential_templates)
                template = found_template.template.name
            except TemplateDoesNotExist:
                raise Http404(f"{slug} cannot be found") from None

            context.update(
                {
                    "data": get_data(),
                    "slug": slug,
                    "template": template,
                    "now": now(),
                }
            )

    if set_in_cache:
        _set_in_cache_if_enabled(slug, template, context)

    response = render(
        request,
        template,
        context=context,
    )

    view_cache = ViewCache()

    if view_cache.is_enabled:
        patch_response_headers(response, cache_timeout=view_cache.seconds)

    return response


def file(request, file_name: str) -> FileResponse:  # noqa: ARG001
    """
    Serves a file from the file path.
    Based on code in https://adamj.eu/tech/2022/01/18/how-to-add-a-favicon-to-your-django-site/#what-the-file-type.
    """

    file_path = get_file_path(file_name)

    if not file_path.exists():
        raise Http404(f"{file_path.name} cannot be found")

    # Don't use context manager to open the file because it will be closed automatically
    # per https://docs.djangoproject.com/en/4.0/ref/request-response/#fileresponse-objects
    file = file_path.open("rb")

    return FileResponse(file)



import json
import logging
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Iterable, Optional

from coltrane.config.cache import DataCache
from coltrane.config.paths import get_content_directory, get_data_directory, get_data_json
from coltrane.config.settings import get_data_json_5
from coltrane.utils import dict_merge

logger = logging.getLogger(__name__)


MARKDOWN_EXTENSION_LENGTH = 3


def _add_data_from_path(data, data_directory, path):
    if path.is_file():
        directory_without_base_and_file_name = (str(path)).replace(str(data_directory), "").replace(path.name, "")

        # TODO: Check that .json5/.json is an extension first
        file_name = path.name.replace(".json5", "").replace(".json", "")
        value = None

        if get_data_json_5():
            try:
                import pyjson5

                try:
                    value = pyjson5.decode_buffer(path.read_bytes())
                except pyjson5.Json5DecoderException:
                    logger.exception(f"Invalid JSON5: '{file_name}'")
            except ImportError:
                pass
        else:
            try:
                value = json.loads(path.read_bytes())
            except json.decoder.JSONDecodeError:
                logger.exception(f"Invalid JSON: '{file_name}'")

        if value:
            new_data = {file_name: value}

            # For each part of the path between BASE_DIR/data and the JSON file,
            # add a new level (i.e. key) in the data dictionary; for example:
            # base_dir/data/some/new/test/here.json with {"one": "two"} ==
            # {"some": {"new": {"test": {"here": {"one": "two"}}}}}
            for key in reversed(directory_without_base_and_file_name.split("/")):
                if key:
                    new_data = {key: new_data}

            data = dict_merge(data, new_data)


def get_data() -> Dict:
    """
    Get and merge data from `data.json` and any JSON files recursively found in the
    `data` directory.
    """

    data = {}
    data_cache = DataCache()
    cache_key = ""

    if data_cache.is_enabled:
        cache_key = f"{data_cache.cache_key_namespace}data"
        data = data_cache.cache.get(cache_key, {})

        if data:
            return data

    if get_data_json().exists():
        warnings.warn("Reading data.json is deprecated; use the data directory instead.", stacklevel=1)

    data_directory = get_data_directory()

    for path in data_directory.rglob("*.json5"):
        _add_data_from_path(data, data_directory, path)

    for path in data_directory.rglob("*.json"):
        _add_data_from_path(data, data_directory, path)

    if data_cache.is_enabled:
        data_cache.cache.set(cache_key, data, timeout=data_cache.seconds)

    return data


def get_content_paths(slug: Optional[str] = None) -> Iterable[Path]:
    """
    Yield `Path`s for all markdown content in the content directory.
    """

    directory = get_content_directory()

    if slug:
        directory = directory / slug

    if not directory.exists():
        raise FileNotFoundError(f"Directory does not exist: {directory}")

    paths = directory.rglob("*.md")

    for path in paths:
        if path.is_file():
            yield path


@dataclass
class ContentItem:
    path: Path
    metadata: Dict
    relative_url: str
    html: str


def get_content_items(skip_draft: bool = True) -> Iterable[ContentItem]:  # noqa: FBT001, FBT002
    from coltrane.renderer import MarkdownRenderer

    paths = get_content_paths()
    _items = []

    content_directory = get_content_directory()
    content_directory_path_length = len(str(content_directory))

    for path in paths:
        (html, metadata) = MarkdownRenderer.instance().render_markdown_path(path)

        if skip_draft and metadata and "draft" in metadata and metadata["draft"] is True:
            continue

        path_str = str(path)
        relative_url = path_str[content_directory_path_length:-MARKDOWN_EXTENSION_LENGTH]

        if relative_url.endswith("/index"):
            relative_url = relative_url[:-6]

        content_item = ContentItem(path=path, metadata=metadata, html=html, relative_url=relative_url)
        _items.append(content_item)

    return _items



#!/usr/bin/env python

# Generated by coltrane==__coltrane_version__

from django.core.management import execute_from_command_line

from coltrane import initialize

wsgi = initialize()

if __name__ == "__main__":
    execute_from_command_line()



"""
A little wrapper CLI over a few Django management commands.
"""

from importlib.metadata import version
from os import getcwd
from pathlib import Path
from stat import S_IEXEC
from subprocess import run as run_process

import click
import rich_click  # type: ignore
from django.core.management.utils import get_random_secret_key

click.Command.format_help = rich_click.rich_format_help  # type: ignore
click.Group.format_help = rich_click.rich_format_help  # type: ignore


FILES_PATH = Path(__file__).parent / "default-files"


def _run_management_command(command_name, *args):
    current_directory = getcwd()
    file_path = Path(current_directory) / "app.py"

    run_process([file_path, command_name, *list(args)])  # noqa: S603, PLW1510


def _copy_file(file_name) -> None:
    default_file_name = file_name

    if file_name.startswith("."):
        default_file_name = file_name[1:]

    with (FILES_PATH / default_file_name).open() as f:
        file_text = f.read()

        if "__coltrane_version__" in file_text:
            coltrane_version = version("coltrane")
            file_text = file_text.replace("__coltrane_version__", coltrane_version)

        if "__app_name__" in file_text:
            current_directory = getcwd().split("/")
            app_name = current_directory[len(current_directory) - 1 :][0]

            file_text = file_text.replace("__app_name__", app_name)

        Path(file_name).write_text(file_text)


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        try:
            cmd_name = ALIASES[cmd_name].name
        except KeyError:
            pass
        return super().get_command(ctx, cmd_name)


@click.group(cls=AliasedGroup, help="Runs commands for coltrane.")
@click.version_option()
def cli():
    pass


@click.command(help="Create a new coltrane site.")
@click.option("--force/--no-force", default=False, help="Force creating a new site")
def create(force):
    app_file = Path("app.py")

    if app_file.exists() and not force:
        click.secho("Skip project creation because app.py already exists.", fg="red")
    else:
        click.secho("Start creating files...\n", fg="yellow")
        Path("__init__.py").touch()

        click.secho("- Create app.py")
        _copy_file("app.py")

        click.secho("- Set app.py as executable")
        app_file.chmod(app_file.stat().st_mode | S_IEXEC)

        click.secho("- Create .env")
        _copy_file(".env")

        # Add randomly generated secret key
        with Path(".env").open("a") as f:
            f.write(f"SECRET_KEY={get_random_secret_key()}")

        click.secho("- Create .watchmanconfig")
        _copy_file(".watchmanconfig")

        click.secho("- Create .gitignore")
        _copy_file(".gitignore")

        click.secho("- Create README.md")
        _copy_file("README.md")

        click.secho("- Create gunicorn.conf.py")
        _copy_file("gunicorn.conf.py")

        click.secho("- Create Dockerfile")
        _copy_file("Dockerfile")

        click.secho("- Create content directory")
        Path("content").mkdir(exist_ok=True)
        (Path("content") / "index.md").write_text("# index.md")

        click.secho("- Create data directory")
        Path("data").mkdir(exist_ok=True)

        click.secho("- Create static directory")
        Path("static").mkdir(exist_ok=True)

        click.secho("- Create templates directory")
        Path("templates").mkdir(exist_ok=True)

    click.secho()
    click.secho("For local development: ", nl=False, fg="green")
    click.secho("poetry run coltrane play")
    click.secho("Build static HTML: ", nl=False, fg="green")
    click.secho("poetry run coltrane record")


@click.command(help="Start a local development server.")
@click.option("--port", default=8000, help="Port to serve localhost from")
def play(port):
    _run_management_command("runserver", f"0:{port}")


@click.command(help="Generates HTML output.")
@click.option("--force/--no-force", default=False, help="Force HTML generation")
@click.option("--threads", type=int, help="Number of threads to use when generating static files")
@click.option("--output", help="Output directory")
@click.option("--ignore/--no-ignore", default=False, help="Ignore errors")
def record(force, threads, output, ignore):
    args = []

    if force:
        args.append("--force")

    if output:
        args.append("--output")
        args.append(output)

    if threads:
        args.append("--threads")
        args.append(str(threads))

    if ignore:
        args.append("--ignore")

    _run_management_command("build", *args)


cli.add_command(create)
cli.add_command(play)
cli.add_command(record)


ALIASES = {
    "init": create,
    "serve": play,
    "build": record,
    "rec": record,
}



from django.contrib.sitemaps import Sitemap

from coltrane.retriever import ContentItem, get_content_items


class ContentSitemap(Sitemap):
    changefreq = "hourly"
    priority = 0.5

    def items(self):
        return get_content_items()

    def location(self, content_item: ContentItem) -> str:
        return content_item.relative_url
